import numpy as np
import dateparser
import datetime

def simple_confirm_value_in_range(field_name, value, min_value, max_value):
  if value is None or value == '':
    return
  value = float(value)
  if value < min_value or value > max_value:
    raise ValueError(f"Invalid value for {field_name}: {value}")
  return value

# interpret whether a value should be None
# which is none, numpy nan, empty string, etc
def should_value_be_none(value):
  if value is None:
    return None
  if isinstance(value, str) and value == '':
    return None
  if isinstance(value, float) and np.isnan(value):
    return None
  return value

def validate_dt_or_convert_to_datetime_obj(dt):
  if isinstance(dt, datetime.datetime):
    return dt
  elif isinstance(dt, str):
    return dateparser.parse(dt)
  else:
    raise ValueError(f"Invalid datetime object: {dt}")



from .weather_station_test_case import WeatherStationTestCase
from .collection_test_case import CollectionTestCase
from .file_storage_test_case import FileStorageTestCase
from .raster_band_test_case import RasterBandTestCase
from .wind_rose_test_case import WindRoseTestCase
from .kriger_test_case import KrigerTestCase
from .tow_calculator_test_case import TOWCalculatorTestCase
from .vector_layer_test_case import VectorLayerTestCase

# algorithm tests
from .grid_to_point_test_case import GridToPointTestCase
from .raster_distance_to_vector_test_case import RasterDistanceToVectorTestCase



import numpy as np
from .helpers import should_value_be_none, simple_confirm_value_in_range, \
  validate_dt_or_convert_to_datetime_obj
from windrose import WindroseAxes
import matplotlib.pyplot as plt

class WindRose:

  def __init__(self, bins=4, precision=2):
    self.bins = bins
    if bins == 4:
      self.directions = ['N', 'E', 'S', 'W']
      self.angles = [x for x in range(0, 360, 90)]
    elif bins == 8:
      self.directions = ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']
      self.angles = [x for x in range(0, 360, 45)]
    elif bins == 16:
      self.directions = ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE',
        'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW']
      self.angles = [x/10.0 for x in range(0, 3600, 225)]
    else:
      raise ValueError("Only 4 or 8 bins are supported")

    # add a CALM bucket for winds that are calm
    self.directions.append('CALM')
    self.calm_threshold_m_s = 0.5 

    self._precision = precision
    self.wind_vectors = {}
    self.wind_values = {}

    self.wind_speed_values = {}
    self.wind_bearing_values = {}

  def direction_to_angle(self, direction):
    return self.angles[self.directions.index(direction)]

  def get_wind_vectors(self):
    return self.wind_vectors

  def get_wind_values(self):
    return self.wind_values

  def get_wind_speed_values(self):
    return self.wind_speed_values

  def get_wind_bearing_values(self):
    return self.wind_bearing_values

  def add_wind_speed(self, speed_m_s, dt):
    if speed_m_s == '':
      speed_m_s = None
    if speed_m_s is not None:
      speed_m_s = round(float(speed_m_s), self._precision)
      speed_m_s = simple_confirm_value_in_range('speed_m_s', speed_m_s, 0, 100)
    self.wind_speed_values[dt] = speed_m_s

  def add_wind_bearing(self, bearing, dt):
    if bearing == '':
      bearing = None
    if bearing is not None:
      bearing = round(float(bearing), self._precision)
      if bearing < 0:
        bearing += 360

      # validate bearing between 0 and 360
      simple_confirm_value_in_range('bearing', bearing, 0, 360)

    self.wind_bearing_values[dt] = bearing

  # output should look like: {'percent': 0.0, 'mean_wind_speed': None}
  # where percent is the % of the total values that are in the bin
  # and mean_wind_speed is the mean wind speed for that bin
  def get_wind_rose(self):
    result = {direction: {'x': 0, 'y': 0, 'count': 0} for direction in self.directions}

    bin_size = 360 / self.bins
    for (dt, (speed, bearing)) in self.wind_values.items():
      (x, y) = self._wind_to_vector(bearing, speed)

      # if the wind is calm, then just add it to the CALM bucket
      if speed and speed < self.calm_threshold_m_s:
        direction = 'CALM'
      else:
        direction = self.bearing_to_direction(bearing)

      result[direction]['x'] += x
      result[direction]['y'] += y
      result[direction]['count'] += 1

    for direction in result.keys():
      if result[direction]['count'] > 0:
        mean_x = result[direction]['x'] / result[direction]['count']
        mean_y = result[direction]['y'] / result[direction]['count']
        mean_speed = round(np.sqrt(mean_x**2 + mean_y**2), self._precision)
        percent = result[direction]['count'] / len(self.wind_values)
        percent *= 100 # convert to percentage
        result[direction] = {'mean_wind_speed': mean_speed,
          'percent': round(percent, self._precision)}
      else:
        result[direction] = {'mean_wind_speed': None, 'percent': 0.0}
    return result

  # three step process
  # 1. find the unique pairs of speed, bearing dt values
  # 2. see which ones are NOT in wind_vectors
  # 3. call add_wind for those vectors
  # do this in an O(1) fashion
  def recalculate_wind_vectors(self):
    calculated_dts = self.wind_vectors.keys()
    not_calculated_dts = set(self.wind_speed_values.keys()) - set(calculated_dts)
    for dt in not_calculated_dts:
      speed = self.wind_speed_values[dt]
      if speed is None:
        continue
      if dt in self.wind_bearing_values.keys():
        bearing = self.wind_bearing_values[dt]
        if bearing is None:
          continue
        self.add_wind(speed, bearing, dt, False)

  def add_wind(self, speed, bearing, dt, add_values=True):
    if speed is None or bearing is None:
      return

    bearing = float(bearing)
    if bearing < 0:
      bearing += 360
    bearing = int(bearing)
    speed = float(speed)
    self.wind_vectors[dt] = self._wind_to_vector(bearing, speed)
    self.wind_values[dt] = (speed, bearing)
    if add_values == True:
      self.wind_speed_values[dt] = speed
      self.wind_bearing_values[dt] = bearing

  def _wind_to_vector(self, bearing, speed):
    if speed is None or bearing is None:
      return None
    bearing_rad = np.radians(bearing)
    x = speed * np.sin(bearing_rad)
    y = speed * np.cos(bearing_rad)
    return (x, y)

  def bearing_to_direction(self, bearing):
    if self.bins == 4:
      index = int((bearing + 45) // 90)
      return self.directions[index % 4]
    elif self.bins == 8:
      directions = ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']
      index = int((bearing + 22.5) // 45)
      return self.directions[index % 8]
    elif self.bins == 16:
      directions = ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE',
        'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW']
      index = int((bearing + 11.25) // 22.5)
      return self.directions[index % 16]
    else:
      raise ValueError("Only 4, 8, or 16 bins are supported")

  def get_wind(self, measure='VECTOR_MEAN'):
    self.recalculate_wind_vectors()
    measure = measure.upper()
    if measure == 'VECTOR_MEAN':
      total_x = 0
      total_y = 0
      count = 0
      for dt, (x, y) in self.wind_vectors.items():
        total_x += x
        total_y += y
        count += 1
      if count == 0:
        return (None, None, None)
      avg_speed = np.sqrt(total_x**2 + total_y**2) / count
      bearing_rad = np.arctan2(total_x, total_y)
      bearing_deg = np.degrees(bearing_rad)
      dir_string = self.bearing_to_direction(bearing_deg)
      if bearing_deg < 0:
        bearing_deg += 360
      return (round(avg_speed, self._precision),
        round(bearing_deg, self._precision), dir_string)
    else:
      raise ValueError(f"Invalid measure: {measure}")

  # plot the wind rose
  def plot(self):
    wd = np.array(list(self.get_wind_bearing_values().values()))
    ws = np.array(list(self.get_wind_speed_values().values()))
    ax = WindroseAxes.from_ax()
    ax.bar(wd, ws, normed=True, opening=0.8, edgecolor="white", blowto=True, nsector=self.bins)
    ax.set_legend()
    return ax



# RasterDistanceToVector
# this algorith is primarily for taking an empty
# raster band and computing the distance to the 
# nearest vector object 
# 
# Example: Raster Distance to Ocean
# Example: Raster Distance to Roads

from ..raster_band import RasterBand
from ..vector_layer import VectorLayer
from osgeo import ogr, osr
import numpy as np
import logging
import os
from multiprocessing import Pool

logger = logging.getLogger(__name__)

# for each row, we're going to want to open up the 
# vector file in read mode, filter it and then find
# distances
def calculate_row(row_data, vector_serialized, buffer_dist):
  vector = VectorLayer()
  vector.deserialize(vector_serialized)
  row_output = []
  for (center_pt, col) in row_data:
    center_geom = ogr.Geometry(ogr.wkbPoint)
    center_geom.AddPoint_2D(center_pt[0], center_pt[1])

    # get the distance to the nearest vector object
    (nearest, dist) = vector.find_nearest_feature(center_geom, buffer_dist)
    if nearest is None:
      row_output.append(np.nan)
    else:
      row_output.append(dist)
  return row_output

class RasterDistanceToVector:

  def __init__(self, raster_band_object, cpus=None):
    assert isinstance(raster_band_object, RasterBand)
    self.raster_band = raster_band_object
    self.cpus = cpus
    if cpus is None:
      self.cpus = os.cpu_count()
    logger.info(f"using {self.cpus} cpus")

  def calculate_distances(self, vector_layer, buffer_dist):
    # check if vector_layer is an instance of VectorLayer
    assert isinstance(vector_layer, VectorLayer), "vector_layer must be an instance of VectorLayer"

    # prepare the vector layer in memory so it goes faster
    logger.info("Memoizing vector layer in memory..")
    memoized = vector_layer.memoize()
    logger.info("Serializing memoized vector layer..")
    serialized = memoized.serialize(True)

    logger.info("Calculating distances to vector objects")
    with Pool(processes=self.cpus) as pool:
      # for reach row in the grid, we will pass it to another function
      # that will calculate the distance to the nearest vector object
      # and return the distance value
      total_rows = self.raster_band.height()
      logger.info(f"Calculating distances for {total_rows} rows")
      pool_of_rows = []

      for row in self.raster_band.rows(True):
        result_row = pool.apply_async(calculate_row, args=(row, serialized, buffer_dist, ))
        pool_of_rows.append(result_row)
        if len(pool_of_rows) % 100 == 0:
          logger.info(f"Putting {len(pool_of_rows)}/{total_rows} rows into queue")

      # create a new band and we're going to add one row at a time into this band
      logger.info(f"Creating new raster band, to populate")
      b2 = self.raster_band.clone_with_no_data()
      count = 0
      for row in pool_of_rows:
        try:
          result = row.get()
        except Exception as e:
          logger.error(f"Error calculating row: {e}")
          raise e
        b2.add_row(count, result)
        count += 1
        if count % 100 == 0:
          logger.info(f"Processed {count}/{total_rows} rows")

      return b2



import logging
import datetime
import os
import numpy as np
from pyproj import CRS
from osgeo import gdal, ogr, osr
from .file_storage import FileStorage

logger = logging.getLogger(__name__)

MOLLWEIDE = '+proj=moll +lon_0=0 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs +type=crs'

class RasterBand:

  def __init__(self, projection=None):
    self._tif = None
    self._band = None
    self._nodata = None
    self._cols = None
    self._rows = None
    self._x_origin = None
    self._y_origin = None
    self._pixel_w = None
    self._pixel_h = None
    self._storage_method = None
    self._extent = None
    self._metadata = None

    # projection stuff 
    self._datum = None
    self._epsg_code = None
    self._proj4_string = None
    self._projection_wkt = None

  def load_url(self, file_url, md5_hash=None):
    logger.debug("opening %s" % file_url)
    fs = FileStorage()
    fs.set_file_url(file_url)
    fs.download()
    self.loadf(fs.get_full_path_to_file())
  
  # this clones the raster band, returning a new 
  # raster band (using load array) but with the 
  # current ones parameters like transforms, etc
  def clone_with_new_data(self, nparray=None):
    if nparray is not None: # if we have new data
      assert nparray.shape[0] == self._rows, "Rows do not match"
      assert nparray.shape[1] == self._cols, "Cols do not match"
    new_file = RasterBand()
    new_file.blank_raster(self._rows, self._cols, 
      (self._pixel_w, self._pixel_h),
      (self._x_origin, self._y_origin))

    if nparray is None:
      nparray = np.full((self._rows, self._cols), self._nodata)

    new_file.load_array(nparray, self._nodata)
    new_file.set_projection(self._projection_wkt)
    return new_file

  def clone_with_no_data(self):
    if self.get_nodata() is None:
      raise ValueError("No nodata set on parent")
    return self.clone_with_new_data(None)

  def get_datum(self):
    return str(self._datum)

  def get_projection(self):
    return self._projection_wkt

  def get_projection_wkt(self):
    return self.get_projection()

  def get_projection_epsg(self):
    projection_wkt = self.get_projection()
    if projection_wkt is None:
      return
    crs = CRS.from_wkt(projection_wkt)
    return crs.to_epsg()

  def get_projection_proj4(self):
    projection_wkt = self.get_projection()
    if projection_wkt is None:
      return
    crs = CRS.from_wkt(projection_wkt)
    return crs.to_proj4()

  # sets the projection on the global tif/raster object
  # this is only for EPSG projections, if you want others
  # then you need to call set_projection with full string
  def set_projection_epsg(self, code):
    if code is None:
      logger.warning("Can't set projection to None")
      return
    if code == self._epsg_code:
      logger.warning(f"Projection already set to {epsg_code}")
      return
    crs = CRS.from_epsg(code)
    projection_wkt = crs.to_wkt()
    self.set_projection(projection_wkt)

  def set_projection_proj4(self, proj_string):
    if proj_string is None:
      logger.warning("Can't set projection to None")
      return
    crs = CRS.from_proj4(proj_string)
    projection_wkt = crs.to_wkt()
    self.set_projection(projection_wkt)

  def set_projection(self, projection_wkt):
    srs = osr.SpatialReference()
    srs.ImportFromWkt(projection_wkt)
    self._tif.SetProjection(srs.ExportToWkt())

    # now set the datum, etc
    crs = CRS.from_wkt(projection_wkt)
    self._datum = crs.datum
    self._epsg_code = crs.to_epsg()
    self._proj4_string = crs.to_proj4()
    self._projection_wkt = srs.ExportToWkt()

  # clips a raster map to a vector layer
  # by creating a mask which is the layer and then
  # applying it to the raster
  def clip_to_vector_layer_extent(self, vector_layer):
    self._throw_except_if_band_not_loaded()

    # make sure vector_layer and raster are in the same projection
    vector_layer_epsg = vector_layer.get_projection_epsg()
    raster_epsg = self.get_projection_epsg()
    if raster_epsg and vector_layer_epsg != raster_epsg:
      err = f"Vector layer and raster are in different projections: {vector_layer_epsg} != {raster_epsg}"
      raise ValueError(err)
    else:
      vector_layer_proj = vector_layer.get_projection_proj4()
      raster_proj = self.get_projection_proj4()
      if vector_layer_proj != raster_proj:
        err = f"Vector layer and raster are in different projections: {vector_layer_proj} != {raster_proj}"
        raise ValueError(err)

    extent_dict = vector_layer.get_extent()
    ul = (extent_dict['min_x'], extent_dict['max_y'])
    lr = (extent_dict['max_x'], extent_dict['min_y'])
    return self.clip_to_extent(ul, lr)

  # clips a map to a new extent
  def clip_to_extent(self, ul, lr):
    self._throw_except_if_band_not_loaded()

    (ul_row, ul_col) = self._map_xy_to_rowcol(ul[0], ul[1])
    (lr_row, lr_col) = self._map_xy_to_rowcol(lr[0], lr[1])

    logger.info(f"ul_row={ul_row} ul_col={ul_col} lr_row={lr_row} lr_col={lr_col}")

    if ul_row > lr_row or ul_col > lr_col:
      raise ValueError("Invalid extent")

    # if any values are less than 0 or greater than the rows/cols
    # then we need to adjust them
    ul_row = max(0, ul_row)
    ul_col = max(0, ul_col)
    lr_row = min(self._rows, lr_row)
    lr_col = min(self._cols, lr_col)

    # ok so now grab only elements of the array
    # that are in the range above
    data = self._band.ReadAsArray(ul_col, ul_row, lr_col - ul_col, lr_row - ul_row)
    data = np.where(data == self._nodata, np.nan, data)

    # now create the new clipped raster
    new_raster = RasterBand()
    new_raster.blank_raster(lr_row - ul_row, lr_col - ul_col,
      (self._pixel_w, self._pixel_h),
      (ul[0], ul[1]))
    new_raster.load_array(data)
    new_raster.set_projection(self._projection_wkt)
    return new_raster

  def get_extent(self):
    self._throw_except_if_band_not_loaded()
    return self._extent

  # apply an arbitrary function over all grid cells
  # no idea how efficient this is
  def apply_function(self, func):
    self._throw_except_if_band_not_loaded()
    assert callable(func), "Function must be callable"
    self.load_array(func(self.values()))

  def get_bounds_from_epsg(self, epsg_code):
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg_code)
    area_of_use = srs.GetAreaOfUse()
    if area_of_use is not None:
      min_x = area_of_use.west_lon_degree
      min_y = area_of_use.south_lat_degree
      max_x = area_of_use.east_lon_degree
      max_y = area_of_use.north_lat_degree
      extent = (min_x, min_y, max_x, max_y)
    else:
      raise ValueError(f"Could not determine the extent for EPSG:{epsg_code}")
    return extent

  def reproject_epsg(self, epsg_code, width=None, height=None):
    crs = CRS.from_epsg(epsg_code)
    proj_wkt = crs.to_wkt()
    return self.reproject(proj_wkt, width, height)

  def reproject_proj4(self, proj_string, width=None, height=None):
    crs = CRS.from_proj4(proj_string)
    proj_wkt = crs.to_wkt()
    return self.reproject(proj_wkt, width, height)

  def reproject_mollweide(self, width=None, height=None):
    MOLLWEIDE = '+proj=moll +lon_0=0 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs +type=crs'
    return self.reproject_proj4(MOLLWEIDE, width, height)

  # reproject will return a new raster band in memory
  # that is in memory but has the new projection
  def reproject(self, projection_wkt, width=None, height=None):
    self._throw_except_if_band_not_loaded()
    if not self._tif.GetProjection():
      raise ValueError("No projection set on current map. Cannot reproject")
    to_srs = osr.SpatialReference()
    to_srs.ImportFromWkt(projection_wkt)
    crs = CRS.from_wkt(projection_wkt)
    code = crs.to_epsg()
    if code == 4326:
      to_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    from_srs = osr.SpatialReference()
    from_srs.ImportFromWkt(self._tif.GetProjection())
    if from_srs.GetAuthorityCode(None) == '4326':
      from_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    transform = osr.CoordinateTransformation(from_srs, to_srs)

    params = {'dstSRS': to_srs.ExportToWkt(),
      'format': 'MEM',
      'resampleAlg': gdal.GRA_NearestNeighbour,
      'warpOptions': [
        'NUM_THREADS=ALL_CPUS',
        'CONFIG GDAL_CACHEMAX 10000'
      ]}
    if height is not None:
      params['height'] = height
    if width is not None:
      params['width'] = width
    in_memory_output = gdal.Warp('', self._tif, **params) # MAIN CALL

    # this new map has a new set of geotransforms, etc
    new_pixel_width = abs(in_memory_output.GetGeoTransform()[1])
    new_pixel_height = abs(in_memory_output.GetGeoTransform()[5])
    new_x_origin = in_memory_output.GetGeoTransform()[0]
    new_y_origin = in_memory_output.GetGeoTransform()[3]

    new_raster = RasterBand()
    new_raster.blank_raster(self._rows, self._cols,
      (new_pixel_width, new_pixel_height),
      (new_x_origin, new_y_origin))
    new_raster._tif = in_memory_output
    new_raster.set_projection(projection_wkt)
    new_raster.set_band(1)   
    return new_raster

  def get_pixel_width(self):
    return self._pixel_w

  def get_pixel_height(self):
    return self._pixel_h

  # This is important, it creates a blank in memory raster
  def blank_raster(self, rows, cols, pixel_widths=(1,1), ul=(0, 0), dtype=gdal.GDT_Float32):
    if type(pixel_widths) is int:
      pixel_widths = (pixel_widths, pixel_widths)
    self._tif = (gdal.GetDriverByName('MEM').Create('', cols, rows, 1, dtype))
    self._storage_method = "MEM"
    self._tif.SetGeoTransform((ul[0], pixel_widths[0], 0, ul[1], 0, -pixel_widths[1]))
    self._cols = cols
    self._rows = rows
    self._x_origin = ul[0]
    self._y_origin = ul[1]
    self._pixel_w = pixel_widths[0]
    self._pixel_h = pixel_widths[1]

  def mean(self):
    return np.nanmean(self.values())

  def max(self):
    return np.nanmax(self.values())

  def min(self):
    return np.nanmin(self.values())

  # load a numpy array into the raster band
  def load_array(self, nparray, nodata=-9999):
    if type(nparray) is list:
      nparray = np.array(nparray)

    if len(nparray.shape) != 2:
      raise ValueError("Array must be 2d")

    # if we dont yet have a tif set
    if self._tif is None:
      rows = nparray.shape[0]
      cols = nparray.shape[1]
      self.blank_raster(rows, cols)

    self._band = self._tif.GetRasterBand(1)
    logger.info(f"Writing array of shape {nparray.shape} to band")
    self._band.WriteArray(nparray)

    self.set_band(1) # working band
    if nodata is not None:
      self.set_nodata(nodata)

  # load raster band from a file
  def loadf(self, gtif):
    logger.debug("opening file %s" % gtif)
    if not os.path.exists(gtif):
      raise ValueError(f"File does not exist: {gtif}")
    self._tif = gdal.Open(gtif)
    self._storage_method = "FILE"
  
    # get projection from the file, in particular get the EPSG
    srs = osr.SpatialReference()
    projection_wkt = self._tif.GetProjection()
    srs.ImportFromWkt(projection_wkt)

    # use pyproj to parse this projection
    crs = CRS.from_wkt(projection_wkt)
    datum = crs.datum
    epsg_code = crs.to_epsg()

    # we only work with WGS84 maps for now
    if str(datum) not in ('WGS 84', 'World Geodetic System 1984'):
      raise ValueError("Only WGS84 maps are supported. Got %s" % datum)

    # get epsg and proj string 
    epsg_code = crs.to_epsg()
    proj_string = crs.to_proj4()
    logger.info(f"Loaded raster with Datum: {datum} EPSG: {epsg_code}")
    
    self._datum = datum
    self._epsg_code = epsg_code
    self._proj4_string = proj_string
    self._projection_wkt = projection_wkt

  # number of bands in the whole file
  def band_count(self):
    return self._tif.RasterCount

  # upper left point
  def ul(self):
    return (self._x_origin, self._y_origin)

  def lr(self):
    return (self._x_origin + self._cols * self._pixel_w,
      self._y_origin - self._rows * self._pixel_h)

  def set_band(self, band_id):
    return self.load_band(band_id)

  def load_band(self, band_id=1):
    if band_id > self.band_count():
      raise ValueError(f"Invalid band id: {band_id}")
    self._band = self._tif.GetRasterBand(band_id)
    self._nodata = self._band.GetNoDataValue()
    logger.debug("nodata value = %s" % self._nodata)
    self._cols = self._tif.RasterXSize
    self._rows = self._tif.RasterYSize

    # geotransform is (top left x, w-e pixel resolution, rotation, top left y, rotation, n-s pixel resolution)
    transform = self._tif.GetGeoTransform()
    self._x_origin = transform[0] # top left x
    self._y_origin = transform[3] # top left y
    self._pixel_w = transform[1] # w-e pixel resolution
    self._pixel_h = -transform[5] # n-s pixel resolution, negative because y increases down
    self._we_res = transform[1] # w-e pixel resolution
    self._ns_res = abs(transform[5]) # n-s pixel resolution

    # set the extent of this band from the geotransform in native 
    # extent should be (minx, miny, maxx, maxy)
    self._extent = {'min_x': self._x_origin,
      'min_y': self._y_origin - (self._rows * self._pixel_h),
      'max_x': self._x_origin + (self._cols * self._pixel_w),
      'max_y': self._y_origin}

    # set any metadata from the field
    self._metadata = self._tif.GetMetadata()

  def get_metadata(self):
    return self._metadata

  def set_metadata(self, field, name):
    assert field is not None, "Field must not be None"
    self._metadata[field] = name
    self._tif.SetMetadata(self._metadata) # update underlying file

  def width(self):
    self._throw_except_if_band_not_loaded()
    return self._cols

  def height(self):
    self._throw_except_if_band_not_loaded()
    return self._rows

  # NOTE: if you call numpy.shape attribute on the band, it will
  # return you the shape of the numpy array which is (rows, cols)
  # this is the opposite of the shape method here
  def shape(self):
    return (self.width(), self.height())

  def size(self):
    self._throw_except_if_band_not_loaded()
    return self._cols * self._rows

  def _throw_except_if_band_not_loaded(self):
    if self._band is None:
      raise ValueError("No band loaded")

  # returns the centroid for a given box, where the point
  # is the UL point
  def _centroid(self, x, y):
    center_x = x + (self._pixel_w/2) # we go left to right
    center_y = y - (self._pixel_h/2) # we go from the top down
    return (center_x, center_y)

  def write_to_file(self, output_filename, compress=False, overwrite=True):
    return self.save_to_file(output_filename, compress, overwrite)

  def save_to_file(self, output_filename, compress=False, overwrite=True):
    if os.path.exists(output_filename) and not overwrite:
      raise ValueError(f"File exists: {output_filename}")
    if self._projection_wkt is None:
      raise ValueError("No projection set on raster")
  
    options = []
    if compress:
      options = ['COMPRESS=LZW', 'BIGTIFF=YES']

    driver = gdal.GetDriverByName('GTiff')
    out_raster = driver.Create(output_filename, 
      self._cols, self._rows, 1, gdal.GDT_Float32,
      options=options)
    out_raster.SetGeoTransform((self._x_origin, self._pixel_w, 0, 
      self._y_origin, 0, -self._pixel_h))
    out_raster.SetProjection(self._tif.GetProjection())

    # add some metadata with generation date
    now_dt = datetime.datetime.now().isoformat()
    self.set_metadata('GENERATION_DATE', now_dt)
    out_raster.SetMetadata(self.get_metadata())

    # now write the data to the file
    out_band = out_raster.GetRasterBand(1)
    if self._nodata is not None:
      out_band.SetNoDataValue(self._nodata)
    out_band.WriteArray(self._band.ReadAsArray())
    out_band.FlushCache()

  def add_row(self, row_number, row_data):
    self._throw_except_if_band_not_loaded()
    
    # make sure row has the right width
    if len(row_data) != self._cols:
      raise ValueError(f"Row data does not match width of raster. Got {len(row_data)} expected {self._cols}")

    # make sure row number is in bounds
    if row_number < 0 or row_number >= self._rows:
      raise ValueError("Row number out of bounds")

    # if its a list, convert to an numpy array
    # and reshape
    if type(row_data) is list:
      row_data = np.array(row_data)
      row_data = row_data.reshape(1, -1)

    # then replace the existing data with this new data
    self._band.WriteArray(row_data, 0, row_number)

  def rows(self, return_centroids=False):
    for i in range(self._rows):
      y = self._y_origin - (i * self._pixel_h)
      x = self._x_origin
      if return_centroids is True:
        coords = [self._centroid(x + (j * self._pixel_w), y) \
          for j in range(self._cols)]
      else:
        coords = [(x + (j * self._pixel_w), y) for j in range(self._cols)]
      data = self._band.ReadAsArray(0, i, self._cols, 1)[0]
      data = [d if d != self._nodata else np.nan for d in data]
      yield list(zip(coords, data))

  def get_bbox_polygon(self):
    self._throw_except_if_band_not_loaded()
    bbox = self.get_bbox()
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(bbox[0][0], bbox[0][1])
    ring.AddPoint(bbox[1][0], bbox[0][1])
    ring.AddPoint(bbox[1][0], bbox[1][1])
    ring.AddPoint(bbox[0][0], bbox[1][1])
    ring.AddPoint(bbox[0][0], bbox[0][1])
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    return poly

  # return tupble of (ul, lr) coordinates
  def get_bbox(self):
    self._throw_except_if_band_not_loaded()
    ul = (self._x_origin, self._y_origin)
    lr = (self._x_origin + self._cols * self._pixel_w,
      self._y_origin - self._rows * self._pixel_h)
    return (ul, lr)

  # get the centerpoint of the raster in the coordinate system
  # the raster is in
  def get_center(self):
    self._throw_except_if_band_not_loaded()
    return (self._x_origin + (self._cols * self._pixel_w / 2.0),
      self._y_origin - (self._rows * self._pixel_h / 2.0))

  def get_no_data(self):
    return self.get_nodata()

  # get nodata from the band and the object here
  def get_nodata(self):
    self._throw_except_if_band_not_loaded()
    band_nodata = self._band.GetNoDataValue()
    assert band_nodata == self._nodata, "Nodata values do not match"
    return band_nodata

  # return the percentage of values that are nodata
  def percentage_nodata(self):
    self._throw_except_if_band_not_loaded()
    data = self.values()
    # count rows that are NOT nan
    nodata_count = np.count_nonzero(np.isnan(data))

    return nodata_count / data.size

  # return the values in the raster as a numpy array
  # note the values shape is rows + cols
  def values(self):
    self._throw_except_if_band_not_loaded()
    data = self._band.ReadAsArray()
    data = np.where(data == self._nodata, np.nan, data)
    return data

  # sum all the values in the array
  def sum(self):
    self._throw_except_if_band_not_loaded()
    return np.nansum(self.values())

  def gradients(self):
    self._throw_except_if_band_not_loaded()
    dx, dy = np.gradient(self.values())
    return dx, dy

  # return the difference between left and right
  def central_diff_gradients(self):
    self._throw_except_if_band_not_loaded()
    arr = self.values()
    grad_x = np.zeros_like(arr, dtype=float)
    grad_y = np.zeros_like(arr, dtype=float)
    grad_x = np.roll(arr, -1, axis=1) - np.roll(arr, 1, axis=1)
    grad_y = np.roll(arr, -1, axis=0) - np.roll(arr, 1, axis=0)
    return grad_x, grad_y

  # x_res = x resolution, y_res = y resolution
  # needs to be in the same units as the raster
  def central_diff_slopes(self, x_res=None, y_res=None):
    if x_res is None:
      x_res = self._pixel_w
    if y_res is None:
      y_res = self._pixel_h
    logger.info(f"Calculating slopes with x_res={x_res} y_res={y_res}")
    self._throw_except_if_band_not_loaded()
    grad_x, grad_y = self.central_diff_gradients()
    # tan theta = opposite / adj
    slope_x = np.degrees(np.arctan(grad_x / (2 * x_res)))
    slope_y = np.degrees(np.arctan(grad_y / (2 * y_res)))
    return (slope_x, slope_y)
  
  # this one computes the bearing, where
  # N=0, E=90, S=180, W=270
  # it needs to use the gradients first to get the 
  # delta x, delta y
  # keep in mind that with most projections Y is going down
  def central_diff_face_bearing(self):
    self._throw_except_if_band_not_loaded()
    delta_x, delta_y = self.central_diff_gradients() 
    
    # create a mask of cases where both delta x and delta y are not zero
    bearings = np.full(delta_x.shape, np.nan)
    non_zero_mask = (delta_x != 0) | (delta_y != 0)

    # compute the bearings only for elements masked
    # note that arctan2 a positive result is counterclockwise
    # and a negative result is clockwise
    bearing_rad = np.arctan2(-delta_x[non_zero_mask], delta_y[non_zero_mask])
    bearing_deg = np.degrees(bearing_rad) 
    bearing_deg = np.where(bearing_deg < 0, bearing_deg + 360, bearing_deg)
    bearings[non_zero_mask] = bearing_deg

    return bearings

  # override the nodata value on everything
  def set_nodata(self, nodata):
    self._throw_except_if_band_not_loaded()

    if self._band.GetNoDataValue() is not None:
      if nodata != self._band.GetNoDataValue():
        logger.warning("NODATA value already set to %s" % self._band.GetNoDataValue())

    self._nodata = nodata
    if nodata is None:
      result_code = self._band.DeleteNoDataValue()
    else:
      result_code = self._band.SetNoDataValue(nodata)
    if result_code != 0:
      raise ValueError("Could not set nodata value")
    self._band.FlushCache()
    assert nodata == self._band.GetNoDataValue(), "NODATA not set"

  # note that this takes row, col notation, which is the standard
  # notation for this kind of stuff versus x,y
  def get_grid_value(self, row, col):
    self._throw_except_if_band_not_loaded()
    return self._band.ReadAsArray(col, row, 1, 1).tolist()[0][0]

  # this helper function takes an x,y coordate and returns
  # which row and column that it maps to, this is useful for 
  # value lookups or if we're doing some kind of extent clipping
  def _map_xy_to_rowcol(self, x, y):
    row = int(np.floor((self._y_origin - y) / self._pixel_h))
    col = int(np.floor((x - self._x_origin) / self._pixel_w))
    return row, col

  # retrieve a single value at a location
  # the x,y values are in the coordinate system of the raster
  def get_value(self, x, y):
    self._throw_except_if_band_not_loaded()

    point = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(x, y)

    # get the bbox and determine if the point is inside of it
    # by using gdal functions + intersection
    bbox_polygon = self.get_bbox_polygon()
    if not bbox_polygon.Contains(point):
      raise ValueError(f"Point outside raster: {x}, {y}")

    # now we need to figure out which row and column we are in
    # make sure to consider resolution and negative y axis
    row, col = self._map_xy_to_rowcol(x, y)
    data = self._band.ReadAsArray(col, row, 1, 1).tolist()[0][0]

    if data == self._nodata:
      return None

    return data



import gstools as gs
import io 
import numpy as np
import matplotlib.pyplot as plt
from osgeo import gdal

MODEL_TYPES = ['Exponential', 'Gaussian', 'Matern', 'Spherical']

# this class implements gis kriging
# which can take a set of points and interpolate
class Kriger:

  def __init__(self, ul, lr, cell_size):
    self.min_x = ul[0]
    self.min_y = lr[1]
    self.max_x = lr[0]
    self.max_y = ul[1]
    self.cell_size = cell_size
    self.variogram_var = 1
    self._interpolated_values = None
    self._interpolated_variance = None

    # if ul and lr are two dimensional then 2d else 3d
    if len(ul) == 2:
      self.dims = 2
    elif len(ul) == 3:
      self.min_z = ul[2]
      self.max_z = lr[2]
      self.dims = 3
    else:
      raise ValueError("Invalid dimensions for ul and lr")

  # data MUST be a N x 3 list or numpy array
  def set_data(self, data):
    if type(data) == list:
      data = np.array(data)

    if self.dims == 2:
      assert data.shape[1] == 3
    elif self.dims == 3:
      assert data.shape[1] == 4

    # make sure all data points are within the bounds
    assert np.all(data[:, 0] >= self.min_x), "Data point outside of bounds"
    assert np.all(data[:, 0] <= self.max_x), "Data point outside of bounds"
    assert np.all(data[:, 1] >= self.min_y), "Data point outside of bounds"
    assert np.all(data[:, 1] <= self.max_y), "Data point outside of bounds"

    # calculate the variogram variance
    total_variance = np.var(data[:, 2])
    if total_variance > 0:
      self.variogram_var = total_variance

    self.data = data

  # main interpolation function
  # model = Variogram model to use
  #   options = ['Exponential', 'Gaussian', 'Matern', 'Spherical']
  # var = Variogram variance 
  # len_scale = The len_scale (length scale) in a variogram model represents 
  #   the distance over which the spatial correlation between data 
  #   points significantly decreases. In simpler terms, it is the 
  #   distance beyond which the values of the spatial 
  #   variable become essentially uncorrelate
  # nugget = Variogram nugget
  def interpolate(self, model='Exponential', len_scale=1, nodata=False, nugget=0):
    assert model in MODEL_TYPES, "Invalid model type"

    model_class = getattr(gs, model) # variogram model
    model = model_class(dim=self.dims, var=self.variogram_var, len_scale=len_scale, nugget=nugget)

    # make sure data has the right shape
    if self.dims == 2:
      assert self.data.shape[1] == 3
      lons, lats, data = self.data[:, 0], self.data[:, 1], self.data[:, 2]
      cond_pos = [lons, lats]
    elif self.dims == 3:
      assert self.data.shape[1] == 4
      lons, lats, elev, data = self.data[:, 0], self.data[:, 1], self.data[:, 2], self.data[:, 3]
      cond_pos = [lons, lats, elev]
      
    # Set up the kriging "model", which is the variogram model and the data
    krige = gs.krige.Ordinary(model, cond_pos=cond_pos, cond_val=data)
      
    # Create a meshgrid of points to interpolate
    x = np.linspace(self.min_x, self.max_x,
      int((self.max_x - self.min_x) / self.cell_size) + 1)
    y = np.linspace(self.min_y, self.max_y,
      int((self.max_y - self.min_y) / self.cell_size) + 1)

    if self.dims == 2:
      grid_x, grid_y = np.meshgrid(x, y, indexing='xy')
      krige_result = krige([grid_x.flatten(), grid_y.flatten()])
    elif self.dims == 3:
      z = np.linspace(self.min_z, self.max_z,
        int((self.max_z - self.min_z) / self.cell_size) + 1)
      grid_x, grid_y, grid_z = np.meshgrid(x, y, z, indexing='xy')
      krige_result = krige([grid_x.flatten(), grid_y.flatten(), grid_z.flatten()])

    # Perform kriging on the entire grid and extract the values and
    # the variance profile. then take the interpolated values and reshape
    # them to the shape of the grid
    (interpolated_values, krige_variance) = krige_result

    if nodata is True:
      min_variance = np.min(krige_variance)
      max_variance = np.max(krige_variance)
      if min_variance < max_variance:
        normalized_variance = (krige_variance - min_variance) / (max_variance - min_variance)
      interpolated_values[normalized_variance > 0.95] = np.nan

    # now we have to flip the interpolated values so that the 
    # origin is in the upper left
    if self.dims == 2:
      self._interpolated_values = interpolated_values.reshape(grid_y.shape)[::-1, :]
    elif self.dims == 3:
      self._interpolated_values = interpolated_values.reshape(grid_z.shape)[::-1, :, :]

    return self._interpolated_values

  def plot_interpolated_grid(self):
    plt.imshow(self._interpolated_values, origin='upper', cmap='viridis')
    plt.colorbar()
    plt.show()

  # geotiff takes the interpolated grid array and returns a bytes object
  # which is the geotiff 
  def geotiff(self, nodata=-9999):
    driver = gdal.GetDriverByName('MEM')
    raster = driver.Create('geotiff',
      self._interpolated_values.shape[1], 
      self._interpolated_values.shape[0], 1, 
      gdal.GDT_Float32)

    # set the geotransform and proj
    raster.SetGeoTransform((self.min_x, self.cell_size, 0, self.max_y, 0, -self.cell_size))
    raster.SetProjection('EPSG:4326')

    # set the nodata value in interpolated values
    out_values = self._interpolated_values.copy()
    out_values[np.isnan(out_values)] = nodata

    # write the array to the raster
    band = raster.GetRasterBand(1)
    band.SetNoDataValue(nodata)
    band.WriteArray(out_values)
    del out_values

    # flush the raster
    band.FlushCache()

    # Use GDAL's virtual memory file system to write to a virtual file
    memfile_path = '/vsimem/temp_geotiff.tiff'
    gdal.Translate(memfile_path, raster, format='GTiff')

    # Read the virtual file into a bytes object
    memfile = gdal.VSIFOpenL(memfile_path, 'rb')
    geotiff_bytes = gdal.VSIFReadL(1, gdal.VSIStatL(memfile_path).size, memfile)
    gdal.VSIFCloseL(memfile)

    # Clean up the virtual file
    gdal.Unlink(memfile_path)

    return geotiff_bytes



# this library is primary for taking 
# a raster file, which is a grid
# and convertint it to a point file, which
# is a vector
#
# 1 raster -> 1 vector
#
from ..raster_band import RasterBand
from ..vector_layer import VectorLayer
from osgeo import ogr, osr
import numpy as np
import logging

logger = logging.getLogger(__name__)

class GridToPoint:

  def __init__(self, raster_band_object):
    assert isinstance(raster_band_object, RasterBand)
    self.raster_band_object = raster_band_object

  # create a new in memory vector layer
  # keep the same SRS as the raster
  # add a field called value, which is the raster value
  def to_vector(self):
    vector_layer = VectorLayer()
    vector_layer.createmem('grid_to_point')
    vector_layer.create_layer('grid_to_point', 'POINT', 
      self.raster_band_object.get_projection_wkt())
    vector_layer.add_field_defn('value', 'float')

    total_rows = self.raster_band_object.height()

    # iterate through cols and make sure we use the centroid
    rows_completed = 0
    for cols in self.raster_band_object.rows(True): # centroid
      for (centroid_points, value) in cols:
        centroid = ogr.Geometry(ogr.wkbPoint)
        centroid.AddPoint_2D(centroid_points[0], centroid_points[1])

        feature = vector_layer.blank_feature()
        feature.SetGeometry(centroid)

        if value is not None and np.isnan(value) == False:
          feature.SetField('value', float(value))
          vector_layer.add_feature(feature)
      rows_completed += 1
      if rows_completed % 100 == 0:
        logger.info('Completed %s/%s rows', rows_completed, total_rows)

    return vector_layer



import os
import json
import shutil
import logging
import numpy as np
from pyproj import CRS
from osgeo import ogr, osr
from .file_storage import FileStorage

logger = logging.getLogger(__name__)

epsg3857 = osr.SpatialReference()
epsg3857.ImportFromEPSG(3857)
epsg4326 = osr.SpatialReference()
epsg4326.ImportFromEPSG(4326)
epsg4326.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

SHAPE_TYPES = ['POINT', 'LINE', 'POLYGON', 'MULTIPOLYGON']
PROJECTIONS = [4326, 3857]

class VectorLayer:

  def __init__(self):
    # basic file input/output stuff
    self._driver = None
    self._driver_name = None
    self._file_path = None
    self._datasource = None
    self._layer = None
    self._fields = {}

    # projection stuff
    self._datum = None
    self._epsg_code = None
    self._proj4_string = None
    self._projection_wkt = None

    self._max_fid = 0

  def get_layer(self):
    return self._layer

  def get_name(self):
    return self._layer.GetName()

  def set_spatial_filter(self, geom):
    self._layer.SetSpatialFilter(geom)

  def reset_spatial_filter(self):
    self._layer.SetSpatialFilter(None)

  def load_url(self, file_url, md5_hash=None):
    logger.debug("opening %s" % file_url)
    fs = FileStorage()
    fs.set_file_url(file_url)
    fs.download()
     
    # if its a zip file, we need to unzip the file
    # and then pick the new path as the file we load
    if fs.is_zip_file() is True:
      fs.unzip()

    return self.loadf(fs.get_full_path_to_file())

  def get_feature_count(self):
    feature_count = self._layer.GetFeatureCount()
    self._layer.ResetReading()
    return feature_count

  # add a feature to the layer, note that we have to do
  # some checking on FIDs
  def add_feature(self, ogr_feature):
    if ogr_feature.GetFID() > self._max_fid:
      self._max_fid = ogr_feature.GetFID()
    else:
      self._max_fid += 1
      ogr_feature.SetFID(self._max_fid)

    self._layer.CreateFeature(ogr_feature)
    return ogr_feature

  def get_extent(self):
    extent = self._layer.GetExtent()
    min_x, max_x, min_y, max_y = extent
    payload = {'min_x': min_x, 'max_x': max_x, 'min_y': min_y, 'max_y': max_y}
    return payload

  # copies an old feature, creates a new one but
  # with the proper layer defn
  def copy_feature(self, ogr_feature):
    old_defn = ogr_feature.GetDefnRef()
    feature_defn = self._layer.GetLayerDefn()
    new_feature = ogr.Feature(feature_def=feature_defn)
    old_geom = ogr_feature.GetGeometryRef()
    new_geom = old_geom.Clone()

    # if the SRS dont match, then we have to reproject
    old_authority = (old_defn.GetGeomFieldDefn(0)
      .GetSpatialRef()
      .GetAttrValue("AUTHORITY", 1))
    new_authority = (feature_defn.GetGeomFieldDefn(0)
      .GetSpatialRef()
      .GetAttrValue("AUTHORITY", 1))
    if old_authority != new_authority:
      new_geom = self.reproject_geom(old_geom.Clone(),
        int(old_authority), int(new_authority))
    new_feature.SetGeometry(new_geom)

    new_field_names = []
    for i in range(feature_defn.GetFieldCount()):
      new_name = feature_defn.GetFieldDefn(i).GetName()
      new_field_names.append(new_name)

    for i in range(old_defn.GetFieldCount()):
      old_name = old_defn.GetFieldDefn(i).GetName()
      if old_name in new_field_names:
        new_feature.SetField(old_name, ogr_feature.GetField(old_name))
    return new_feature

  # reproject a single geometry obj
  def reproject_geom(self, geom, old_epsg, new_epsg=4326):
    new_geom = geom.Clone()
    epsg_from = osr.SpatialReference()
    epsg_from.ImportFromEPSG(old_epsg)
    epsg_to = osr.SpatialReference()
    epsg_to.ImportFromEPSG(new_epsg)
    if old_epsg == 4269:
      epsg_from.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    if new_epsg == 4326:
      epsg_to.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    transform_func = osr.CoordinateTransformation(epsg_from, epsg_to)
    new_geom.Transform(transform_func)
    return new_geom

  def blank_feature(self, geometry=None):
    feature_defn = self._layer.GetLayerDefn()
    feature = ogr.Feature(feature_def=feature_defn)
    if type(geometry) in [list, tuple]:
      geometry = (float(i) for i in geometry)
      point = ogr.Geometry(ogr.wkbPoint)
      point.AddPoint(*geometry)
      feature.SetGeometry(point)
    elif geometry is not None:
      feature.SetGeometry(geometry)
    return feature

  # clones the current object to an in-memory object
  # useful for speeding things up + testing 
  def clone_to_memory(self, old_layer, copy_features=True):
    assert type(old_layer) == VectorLayer, "Invalid layer type"
    mem_layer = VectorLayer()
    mem_layer.createmem(old_layer.get_name())
    mem_layer.copy_blank_layer(old_layer, old_layer.get_name())
    mem_layer.auto_set_projection()
    for feature in old_layer._layer:
      mem_layer.add_feature(feature.Clone())
    return mem_layer

  def reproject_epsg(self, new_epsg):
    crs = CRS.from_epsg(new_epsg)
    return self._reproject(crs.to_wkt())

  def reproject_proj4(self, proj4_string):
    crs = CRS.from_proj4(proj4_string)
    return self._reproject(crs.to_wkt())
  
  # return feature off the stack
  def get_feature(self):
    return self._layer.GetNextFeature()

  def _reproject(self, projection_wkt):
    crs = CRS.from_wkt(projection_wkt)
    projection_wkt = crs.to_wkt()

    mem_layer = VectorLayer()
    mem_layer.createmem(self._layer.GetName())
    mem_layer.copy_blank_layer(self, self._layer.GetName(), projection_wkt)

    source_srs = self._layer.GetSpatialRef()
    target_srs = osr.SpatialReference()
    target_srs.ImportFromWkt(projection_wkt)
    transform = osr.CoordinateTransformation(source_srs, target_srs)

    # now reproject the individual features
    for feature in self._layer:
      geom = feature.GetGeometryRef()
      geom.Transform(transform)
      fid = feature.GetFID()
      feature.SetGeometry(geom)
      feature.SetFID(fid)
      mem_layer.add_feature(feature)

    # GDAL 3.7 will support the below
    # self._layer.SetActiveSRS(target_srs)
    mem_layer.auto_set_projection()
    return mem_layer

  # this will copy all the features from the current
  # layer to a new layer, including all the fields
  def copy_features(self, new_layer):
    assert self.get_projection_wkt() == new_layer.get_projection_wkt(), "Projections do not match"
    old_defn = self._layer.GetLayerDefn()
    new_defn = new_layer.get_layer().GetLayerDefn()
    old_field_names = [old_defn.GetFieldDefn(i).GetName() for i in range(old_defn.GetFieldCount())]
    new_field_names = [new_defn.GetFieldDefn(i).GetName() for i in range(new_defn.GetFieldCount())]
    assert old_field_names == new_field_names, "Field names do not match"

    # then copy the features
    for i in range(self.get_feature_count()):
      old_feature = self.get_feature(i)
      new_feature = new_layer.blank_feature()
      new_feature = self.copy_feature(old_feature)
      new_layer.add_feature(new_feature)

  def createmem(self, name):
    logger.info("creating an in-memory layer %s" % name)
    self._driver = ogr.GetDriverByName('MEMORY')
    self._driver_name = 'MEMORY'
    self._datasource = self._driver.CreateDataSource(name)

  def createf(self, vector_file, overwrite=False):
    logger.info("creating file @ %s" % vector_file)
    self._driver = self.auto_determine_driver(vector_file)
    self._driver_name = self._driver.GetName()
    if overwrite is True:
      if os.path.isfile(vector_file):
        os.remove(vector_file)
      if os.path.isdir(vector_file):
        shutil.rmtree(vector_file)
    self._datasource = driver.CreateDataSource(vector_file)

  def get_driver_name(self):
    return self._driver_name  

  def copy_blank_layer(self, old_vector_layer, new_name, projection_wkt=None):
    logger.info("Copying layer from %s" % old_vector_layer)
    assert self._datasource is not None, "No datasource set. call createf"

    # if projection wkt, use that else use the current layers ref
    if projection_wkt is not None:
      srs = osr.SpatialReference()
      srs.ImportFromWkt(projection_wkt)
    else:
      srs = old_vector_layer._layer.GetSpatialRef()

    geom_type = old_vector_layer._layer.GetGeomType()

    # copy the layer itself and the spatial reference
    old_layer_def = old_vector_layer.get_layer().GetLayerDefn()
    self._layer = self._datasource.CreateLayer(new_name, srs=srs, geom_type=geom_type)

    # copy the field definitions
    field_defs = [old_layer_def.GetFieldDefn(i) for i in range(old_layer_def.GetFieldCount())]
    for field_def in field_defs:
      self._layer.CreateField(field_def)
  
    logger.info("Layer copy completed")

  def create_layer_epsg(self, layer_name, shape_type='POINT', proj=4326):
    crs = CRS.from_epsg(proj)
    return self.create_layer(layer_name, shape_type, crs.to_wkt())

  def create_layer(self, layer_name, shape_type='POINT', projection_wkt=None):
    assert self._datasource is not None, "No datasource set. call createf"
    assert shape_type.upper() in SHAPE_TYPES, "Invalid shape type"
    logging.info("creating layer %s" % layer_name)

    if shape_type.upper() == 'POINT':
      use_shape = ogr.wkbPoint
    elif shape_type.upper() == 'LINE':
      use_shape = ogr.wkbLineString
    elif shape_type.upper() == 'POLYGON':
      use_shape = ogr.wkbPolygon
    elif shape_type.upper() == 'MULTIPOLYGON':
      use_shape = ogr.wkbMultiPolygon

    srs = osr.SpatialReference()
    if projection_wkt is None:
      srs.ImportFromEPSG(4326)
      projection_wkt = srs.ExportToWkt()
    else:
      srs.ImportFromWkt(projection_wkt)

    self._layer = self._datasource.CreateLayer(layer_name,
      srs=srs, geom_type=use_shape)
    self.auto_set_projection()

  def add_field_defn(self, field_name, field_type='int'):
    return self.add_field_def(field_name, field_type)

  def add_field_def(self, field_name, field_type='int'):
    for i in range(self._layer.GetLayerDefn().GetFieldCount()):
      field_def = self._layer.GetLayerDefn().GetFieldDefn(i)
      if field_def.GetName() == field_name:
        logger.info("field with name %s already exists" % field_name)
        return
    if field_name == '':
      logger.info("field name cannot be empty")
      return

    if field_type in (int, 'int', 'Integer', ogr.OFTInteger):
      fd = ogr.FieldDefn(field_name, ogr.OFTInteger)
      fname = 'Integer'
    elif field_type in (float, 'float', 'Real', ogr.OFTReal):
      fd = ogr.FieldDefn(field_name, ogr.OFTReal)
      fname = 'Real'
    elif field_type in (str, 'str', 'String', ogr.OFTString):
      fd = ogr.FieldDefn(field_name, ogr.OFTString)
      fname = 'String'
    else:
      raise ValueError("Invalid field type, must be int, float or str. got %s" % field_type)
    self._fields[field_name] = {'type': fname}
    self._layer.CreateField(fd)

  def get_fields(self):
    return self._fields

  def get_file_path(self):
    return self._file_path

  # memoize - take a layer that is saved on disk
  # and convert it to an in memory layer for faster
  # processing
  def memoize(self):
    if self._driver_name == 'MEMORY':
      raise ValueError("Layer is already in memory")
    mem_layer = VectorLayer()
    mem_layer.createmem(self.get_name())
    mem_layer.copy_blank_layer(self, self.get_name())
    mem_layer.auto_set_projection()
    mem_layer.auto_set_fields()
    for feature in self.get_layer():
      mem_layer.add_feature(feature.Clone())
    return mem_layer

  # materialize - basically the same as saving to file
  # but updates the object so if you do queries against 
  # it, it goes against the layer on disk instead of
  # the layer in memory (can be used to save memory)
  def materialize(self, file_path, overwrite=True):
    self.save_to_file(file_path, overwrite)
    self.loadf(file_path)

  # this will save the layer out to file, which will
  # require opening the and then copying all the features
  def save_to_file(self, file_path, overwrite=True):
    if overwrite is True:
      if os.path.isfile(file_path):
        os.remove(file_path)
      if os.path.isdir(file_path):
        shutil.rmtree(file_path)
    driver = self.auto_determine_driver(file_path)
    logger.info("saving to %s using driver %s" % (file_path, driver.GetName()))
    out_ds = driver.CreateDataSource(file_path)
    out_layer = out_ds.CreateLayer(self._layer.GetName(),
      srs=self._layer.GetSpatialRef(),
      geom_type=self._layer.GetGeomType())
    for i in range(self._layer.GetLayerDefn().GetFieldCount()):
      field_def = self._layer.GetLayerDefn().GetFieldDefn(i)
      field_name = field_def.GetName()
      logger.info("adding field %s" % field_name)
      out_layer.CreateField(field_def)

    save_count = 0
    total_features = self.get_feature_count()
    logger.info("saving %s features to file" % total_features)
    for feature in self._layer:
      geom = feature.GetGeometryRef()
      new_feature = ogr.Feature(out_layer.GetLayerDefn())
      new_feature.SetGeometry(geom)
      for i in range(feature.GetFieldCount()):
        field_name = feature.GetFieldDefnRef(i).GetName()
        field_value = feature.GetField(i)
        new_feature.SetField(field_name, field_value)
      new_feature.SetFID(feature.GetFID())
      out_layer.CreateFeature(new_feature)
      save_count += 1
      if save_count % 1000 == 0:
        logger.info(f"saved {save_count}/{total_features} features")

  # figure out which type of file we are dealing with
  # so we can use the right ogr driver to open it
  def auto_determine_driver(self, filename):
    if 'gdb' in filename:
      return ogr.GetDriverByName('OpenFileGDB')
    elif 'gpkg' in filename:
      return ogr.GetDriverByName('GPKG')
    elif 'shp' in filename:
      return ogr.GetDriverByName('ESRI Shapefile')
    elif 'geojson' in filename:
      return ogr.GetDriverByName('GeoJSON')
    elif 'kml' in filename:
      return ogr.GetDriverByName('KML')
    elif 'vrt' in filename:
      return ogr.GetDriverByName('VRT')
    else:
      raise Exception("Unknown vector format for %s" % filename)

  def loadf(self, vector_file, layer_id=0):
    logger.info("opening %s" % vector_file)
    self._driver = self.auto_determine_driver(vector_file)
    self._driver_name = self._driver.GetName()
    self._source = self._driver.Open(vector_file, 0)
    self._file_path = vector_file
    num_layers = self._source.GetLayerCount()
    logger.info("found %s number of layers" % num_layers)
    if type(layer_id) == int:
      self._layer = self._source.GetLayer(layer_id)
    elif type(layer_id) == str:
      logger.info("fetching layer by name = %s" % layer_id)
      for idx in range(self._source.GetLayerCount()):
        candidate_layer = self._source.GetLayerByIndex(idx)
        candidate_layer_name = candidate_layer.GetName()
        if candidate_layer_name == layer_id:
          logger.info("found layer matching = %s" % idx)
          self._layer = self._source.GetLayer(idx)
    self.auto_set_fields()
    assert self._layer is not None, "cannot find layer id = %s" % layer_id
    self.auto_set_projection()

  # this will automatically look at the fields on the layer
  # and set the fields on this object properly
  def auto_set_fields(self):
    layer_defn = self._layer.GetLayerDefn()
    for i in range(layer_defn.GetFieldCount()):
      field_defn = layer_defn.GetFieldDefn(i)
      field_name = field_defn.GetName()
      field_type = field_defn.GetTypeName()
      self._fields[field_name] = {'type': field_type}

  # this will just set the variables on this object based
  # on what the projection is in the layer
  def auto_set_projection(self):
    projection_wkt = self._layer.GetSpatialRef().ExportToWkt()
    self._projection_wkt = projection_wkt
    crs = CRS.from_wkt(projection_wkt)
    self._datum = crs.datum
    self._epsg_code = crs.to_epsg()
    self._proj4_string = crs.to_proj4()

  def get_datum(self):
    return str(self._datum)

  def get_projection_wkt(self):
    return self._projection_wkt

  def get_projection_epsg(self):
    return self._epsg_code

  def get_projection_proj4(self):
    return self._proj4_string

  # applies an arbitrary function to all features
  def apply_to_features(self, func):
    for feature in self.get_layer():
      func(feature)
      self._layer.SetFeature(feature)

  def find_distance_m(self, geom1, geom2):
    if type(geom1) == tuple and len(geom1) == 2:
      (x, y) = geom1
      tmp_g = ogr.Geometry(ogr.wkbPoint)
      tmp_g.AddPoint_2D(x, y)
      geom1 = tmp_g
    if type(geom2) == tuple and len(geom2) == 2:
      (x, y) = geom2
      tmp_g = ogr.Geometry(ogr.wkbPoint)
      tmp_g.AddPoint_2D(x, y)
      geom2 = tmp_g
    transform_func = osr.CoordinateTransformation(epsg4326, epsg3857)
    geom1.Transform(transform_func)
    geom2.Transform(transform_func)
    return geom1.Distance(geom2)

  # find nearest feature is useful in that it will return the
  # nearest feature, the distance to that feature and the nearest
  # point in that feature
  def find_nearest_feature(self, xy, bounding_shape=None):
    sorted_records = self.find_nearest_features(xy, bounding_shape)
    if len(sorted_records) > 0:
      closest = sorted_records[0][0]
      distance = sorted_records[0][1]
      return (closest, distance)
    return (None, None)

  # returns a list of all features and distances in the 
  # native projection 
  def find_nearest_features(self, xy, bounding_shape=None):
    if type(xy) == tuple and len(xy) == 2:
      (x, y) = xy
      point = ogr.Geometry(ogr.wkbPoint)
      point.AddPoint_2D(x, y)
    elif type(xy) == ogr.Geometry:
      point = xy

    if type(bounding_shape) is float:
      radius = bounding_shape
      bounding_shape = point.Buffer(radius)
    if bounding_shape is not None:
      self._layer.SetSpatialFilter(bounding_shape)
    dists = [(f, point.Distance(f.GetGeometryRef())) for f in self._layer]
    sorted_records = sorted(dists, key=lambda i: i[1])
    return sorted_records

  # method which will serialize the entire object into something
  # that can be deserialized and reloaded. Features is optional and
  # will be included as geojsons
  def serialize(self, include_features=False):
    payload = {
      'projection_wkt': self.get_projection_wkt(),
      'name': self.get_name(),
      'feature_count': self.get_feature_count(),
      'extent': self.get_extent(),
      'fields': self.get_fields(),
      'storage': {
        'driver': self.get_driver_name(),
        'file_path': self.get_file_path()
      }
    }

    is_memory_layer = self.get_driver_name() == 'MEMORY'

    if include_features is True:
      features = []
      for feature in self.get_layer():
        geom = feature.GetGeometryRef()
        geom_json = geom.ExportToJson()
        feature_json = {
          'geometry': geom_json,
          'fields': {feature.GetFieldDefnRef(i).GetName(): feature.GetField(i) for \
              i in range(feature.GetFieldCount())}
        }
        features.append(feature_json)
      payload['features'] = features
    return json.dumps(payload)
  
  # deserialize will load this into memory
  def deserialize(self, payload):
    payload = json.loads(payload)
    projection_wkt = payload['projection_wkt']
    layer_name = payload['name']
    extent = payload['extent']
    fields = payload['fields']

    self.createmem(layer_name)
    self.create_layer(layer_name, 'POINT', projection_wkt)
    logger.info("loading layer %s" % layer_name)
    field_list = []
    for (field_name, field) in fields.items():
      self.add_field_defn(field_name, field['type'])
      field_list.append(field_name)

    self.auto_set_projection()

    # if we have features in here, then create all the features
    if 'features' in payload:
      for feature in payload['features']:
        geom = ogr.CreateGeometryFromJson(feature['geometry'])
        new_feature = self.blank_feature(geom)
        for field_name in field_list:
          field_value = feature['fields'].get(field_name)
          new_feature.SetField(field_name, field_value)
        self.add_feature(new_feature)

    # if we dont have features then we're probably using a
    # layer that is stored on disk
    else:
      self.loadf(payload['storage']['file_path'])

    return self



# file_storage.py
# this is a file storage class which is used to download
# a file from a URL and store it locally
# it also has a method to check the md5 hash of the file
# and if it doesn't match, it will raise an exception
# author: Tom Hayden

import json
import hashlib
import os
import platform
import logging
import datetime
import requests
import zipfile
import shutil

logger = logging.getLogger(__name__)

class FileStorage:

  def __init__(self):
    self.set_cache_dir()
    self.file_url = None
    self.expected_md5_hash = None
    self.relative_path = None

  def get_cache_dir(self):
    return self.cache_dir

  def set_file_url(self, file_url):
    self.file_url = file_url

    # figure out the relative_path from the http url
    self.relative_path = os.path.basename(self.file_url)

  def set_cache_dir(self):
    if platform.system() == 'Windows':
      self.cache_dir = os.getenv('LOCALAPPDATA')
    else:
      self.cache_dir = os.path.expanduser('~/.cache')
    self.cache_dir = os.path.join(self.cache_dir, 'wx_logs')
    if not os.path.exists(self.cache_dir):
      os.makedirs(self.cache_dir)
 
  def get_relative_path_to_file(self):
    return self.relative_path

  def get_full_path_to_file(self):
    return self.get_cache_dir() + '/' + self.get_relative_path_to_file()

  def set_expected_md5_hash(self, md5_hash):
    self.expected_md5_hash = md5_hash

  # return ONLY the relative_path of the file
  def get_file_name(self):
    return self.relative_path

  def get_file_size(self):
    return os.path.getsize(self.get_full_path_to_file())

  # md5 hash of the file
  def get_md5_hash(self):
    file_path = self.get_full_path_to_file()
    with open(file_path, 'rb') as f:
      return hashlib.md5(f.read()).hexdigest()

  # this function looks at the file (or at the contents of the zip file)
  # and determines what kind of file it is
  # supported: GPKG, GTIFF, GDB, SHP, GEOJSON
  def get_gis_file_type(self):
    if self.is_zip_file():
      contents = self.peek_zip_file_toplevel()
      for c in contents:
        if c.endswith('.gpkg'):
          return 'GPKG'
        elif c.endswith('.gdb'):
          return 'GDB'
        elif c.endswith('.shp'):
          return 'SHP'
        elif c.endswith('.geojson'):
          return 'GEOJSON'
        elif c.endswith('.tif') or c.endswith('.tiff'):
          return 'GTIFF'
    else:
      if self.relative_path.endswith('.gpkg'):
        return 'GPKG'
      elif self.relative_path.endswith('.gdb'):
        return 'GDB'
      elif self.relative_path.endswith('.shp'):
        return 'SHP'
      elif self.relative_path.endswith('.geojson'):
        return 'GEOJSON'
      elif self.relative_path.endswith('.tif') or self.relative_path.endswith('.tiff'):
        return 'GTIFF'
    return None

  def is_zip_file(self):
    return self.relative_path.endswith('.zip')

  # if the file is a zip file then peek into it
  # and return the list of contents
  def peek_zip_file(self):
    file_path = self.get_full_path_to_file()
    with zipfile.ZipFile(file_path, 'r') as z:
      return z.namelist()

  # peek into the zipfile and only return the top level
  def peek_zip_file_toplevel(self):
    all_items = self.peek_zip_file()
    top_level = {}
    for item in all_items:
      if '/' in item: # folder
        folder_name = item.split('/')[0]
        if folder_name not in top_level:
          top_level[folder_name] = 1
      else:
        top_level[item] = 1
    return top_level.keys()

  # extract the zip file to the cache directory
  # and return the main file name
  def unzip(self, overwrite=True):
    extract_dir = self.get_cache_dir()
    file_path = self.get_full_path_to_file()
    logger.info(f"Unzipping file: {file_path}")
    with zipfile.ZipFile(file_path, 'r') as z:
      contents = z.namelist()
      if self.zip_needs_subfolder() is False:
        # extract the single file to the cache directory
        z.extractall(extract_dir)
        self._autodetect_new_file(extract_dir)
      else:
        # extract the contents to a subfolder
        subfolder = self.relative_path.replace('.zip', '')
        subfolder = os.path.join(extract_dir, subfolder)
        if os.path.exists(subfolder):
          if overwrite:
            shutil.rmtree(subfolder)
          else:
            return subfolder
        os.makedirs(subfolder)
        z.extractall(subfolder)
        self._autodetect_new_file(subfolder)
  
  # updates the self.relative_path to point to the new
  def _autodetect_new_file(self, new_folder):
    if '.gdb' in new_folder:
      self.relative_path = new_folder
      return True
    if '.gpkg' in new_folder:
      self.relative_path = new_folder
      return True

    # if there is a gdb file in here, then grab that
    folder_contents = os.listdir(new_folder)
    for item in folder_contents:
      if item.endswith('.gdb'):
        new_relative_path = new_folder + '/' + item
        new_relative_path_relative_to_cache = new_relative_path.replace(self.cache_dir + '/', '')
        self.relative_path = new_relative_path_relative_to_cache
        return True

    # if the folder contains a .shp then that .shp file is teh relative_path
    folder_contents = os.listdir(new_folder)
    for item in folder_contents:
      if item.endswith('.shp'):
        new_relative_path = new_folder + '/' + item
        new_relative_path_relative_to_cache = new_relative_path.replace(self.cache_dir + '/', '')
        self.relative_path = new_relative_path_relative_to_cache
        return True

    # if we are here then we dont know what kind of
    # file we're trying to look at so throw exception
    raise ValueError("Could not autodetect new file")

  def zip_needs_subfolder(self):
    file_path = self.get_full_path_to_file()
    with zipfile.ZipFile(file_path, 'r') as z:
      contents = z.namelist()
      return len(contents) > 1

  def delete_file(self):
    file_name = self.get_file_name()
    if os.path.exists(file_name):
      os.remove(file_name)

  # this the main call to download the file, but we should also 
  # make sure to see if it's already there and if so, check the md5 hash
  # and in that case just return bc we are ok
  def download(self, unzip=True):
    if self.file_url is None:
      raise ValueError("No file URL set. call set_file_url()")
    file_name = self.get_full_path_to_file()
    if os.path.exists(file_name):
      if self.expected_md5_hash is not None:
        actual_md5_hash = self.get_md5_hash()
        if actual_md5_hash != self.expected_md5_hash:
          raise ValueError(f"MD5 hash mismatch for file: {self.file_url}")
      return
    logger.info(f"Downloading file: {self.file_url}")
    response = requests.get(self.file_url)
    if response.status_code != 200:
      raise ValueError(f"Invalid response code: {response.status_code}")

    with open(file_name, 'wb') as f:
      f.write(response.content)
    logger.info(f"Downloaded file: {self.file_url}")

    if self.expected_md5_hash is not None:
      actual_md5_hash = self.get_md5_hash()
      if actual_md5_hash != self.expected_md5_hash:
        raise ValueError(f"MD5 hash mismatch for file: {self.file_url}")



import datetime
import numpy as np

TOW_RH_THRESHOLD = 80
TOW_T_THRESHOLD = 0
HOURS_PER_YEAR = 8760

# this class is designed to calculate annual TOW (time of wetness)
# for a given year which is the number of hours in a year that
# the surface is wet. This is important for the calculation of
# corrosion rates for metals.
# number of hours per year with RH > 80% and T > 0
class TOWCalculator:

  def __init__(self, precision=4):
    self._data = {}
    self._precision = precision

  def _validate_dt(self, dt):
    if not isinstance(dt, datetime.datetime):
      raise ValueError('dt must be a datetime object, got %s' % type(dt))
    year = dt.year
    if year not in self._data.keys():
      self.create_empty_year(year)
    day = dt.day
    month = dt.month
    hour = dt.hour
    return (year, month, day, hour)

  def annualize(self, valid_hours, tow_hours):
    if valid_hours == 0:
      return None
    pct_of_valid = float(tow_hours) / float(valid_hours)
    return int(round(pct_of_valid * HOURS_PER_YEAR, 0))

  # return a multi year average, excluding years with QA failures
  def get_averages(self):
    years = self.get_years()
    total_hours = 0
    tow_hours = 0
    max_hours = 0
    valid_years = 0
    for year in years.keys():
      if years[year]['qa_state'] == 'PASS':
        total_hours += years[year]['total_hours']
        tow_hours += years[year]['time_of_wetness_actual']
        max_hours += years[year]['max_hours']
        valid_years += 1

    # ok now do the same projection we do below
    projected_tow = self.annualize(total_hours, tow_hours)

    return {'annual_time_of_wetness': projected_tow,
      'valid_years': valid_years}

  # return an array for each year in data that looks like
  # {'total_hours': N, 'time_of_wetness': N, 'percent_valid': N}
  def get_years(self):
    years = {}
    for year in self._data.keys():
      max_hours = len(self._data[year].keys())
      total_hours = 0
      tow_hours = 0
      for (month, day, hour) in self._data[year].keys():
        temp_readings = self._data[year][(month, day, hour)]['t']
        rh_readings = self._data[year][(month, day, hour)]['rh']
        if len(temp_readings) == 0 or len(rh_readings) == 0:
          continue
        total_hours += 1
        mean_t = np.mean(temp_readings)
        mean_rh = np.mean(rh_readings)
        if mean_t > TOW_T_THRESHOLD and mean_rh > TOW_RH_THRESHOLD:
          tow_hours += 1
      percent_valid = round(float(total_hours) / float(max_hours), self._precision)
      qa_state = 'PASS' if percent_valid > 0.75 else 'FAIL'
      if qa_state == 'PASS':
        projected_tow = self.project_tow(total_hours, max_hours, tow_hours)
      else:
        projected_tow = None
      payload = {'max_hours': max_hours, 
        'total_hours': total_hours,
        'time_of_wetness': projected_tow,
        'time_of_wetness_actual': tow_hours,
        'qa_state': qa_state,
        'percent_valid': percent_valid}
      years[year] = payload
    return years

  # extrapolate total tow based on missing values
  def project_tow(self, hours_with_data, max_hours, current_tow):
    return round((current_tow / hours_with_data) * max_hours, 2)

  # creates an empty year with one row for every hour 
  # of the year
  def create_empty_year(self, year):
    if year not in self._data.keys():
      self._data[year] = {}
    for month in range(1, 13):
      days_in_month = 31
      if month == 2 and year % 4 == 0:
        days_in_month = 29 # leap year
      elif month == 2:
        days_in_month = 28
      elif month in [4, 6, 9, 11]:
        days_in_month = 30
      for day in range(1, days_in_month + 1):
        for hour in range(24):
          self._data[year][(month, day, hour)] = {'t': [], 'rh': []}

  def add_temperature(self, temperature, dt):
    (year, month, day, hour) = self._validate_dt(dt)
    self._data[year][(month, day, hour)]['t'].append(temperature)

  def add_humidity(self, rh, dt):
    (year, month, day, hour) = self._validate_dt(dt)
    self._data[year][(month, day, hour)]['rh'].append(rh)
